#!/usr/bin/env python
# -*- coding: latin-1 -*-

""" The client """

# Imports
import os
import sys
import time
import random
import math
import time

# Pygame imports
import pygame
from pygame.locals import *

# Numarray imports
from numarray import array
from numarray import asarray
from numarray import matrixmultiply
from numarray import arange
from numarray import transpose

# Defined constants
FALSE = 0
TRUE = 1
NULL = 0

MIDDLE = 240
FIRSTELEMENT = 0
SECONDELEMENT = 1
TIME = 0
POINT = 1
X = 0
Y = 1
SIMPLE = 1

UNKNOWN = -1
SAVE = 0
PAINT = 1
TIMES = 2
INPUT = 3
PRED = 4
ERASE = 5

WHITE = (255, 255, 255)
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)
BLACK = (0, 0, 0)

	
# Init modes for the simulation
paintMode = FALSE
timeMode = FALSE
showFilePlot = TRUE
showFilePaintPathPoints = FALSE
showPredictedPlot = TRUE
showExtrapolatedPathPoints = FALSE
showNetworkPoints = FALSE
clearScreen = FALSE
getValue = FALSE


# On/off toggle text
ONOFFTEXT = ["Off", "On"]

# Dictionary for clickable on/off modes
inputDict = { \
	SAVE: FALSE, \
	PAINT: FALSE, \
	TIMES: FALSE, \
	INPUT: FALSE, \
	PRED: FALSE, \
	ERASE: FALSE, \
	UNKNOWN: FALSE, \
	}

DEBUG = FALSE

# Initialize Pygame
pygame.init()
screen = pygame.display.set_mode((800, 600))
pygame.display.set_caption('Prediction simulation')
pygame.mouse.set_visible(1)

# Display some text
pygame.font.init() 
font = pygame.font.Font(pygame.font.match_font("tahoma"), 16)

# Show the background
background = pygame.Surface(screen.get_size())
background = background.convert()
background.fill((250, 250, 250))
screen.blit(background, (0, 0))
pygame.display.flip()



# // ############################# Help methods   #############################
def disableExcept(cmd):
	""" Set every command in inputDict to FALSE """
	for elem in inputDict:
		if elem != cmd:
			inputDict[elem] = FALSE

def getDistance(a, b):
	""" Calculate the distance between a and b """
	result = -1
	distance = asarray(a-b)
	try:
		if len(distance) == 2:
			result = math.sqrt(distance[0]**2 + distance[1]**2)
		elif len(distance) == 1:
			# If parameters are in numarray.array format, len operation is permitted
			result = a[0] - b[0]
	except ValueError:
		# We catch this as an error because we cant make len operation on an interger, float ...
		result = distance
	return abs(result)

def plotText(input, pos=None):
	""" Plot a text at the screen """
	text = font.render(input, 1, (100, 100, 100))
	textpos = text.get_rect()
	textpos.centerx = background.get_rect().centerx
	if pos == None:
		background.blit(text, textpos)
	else:
		background.blit(text, (textpos[0], pos))
	
def plotAverage(input):
	""" Plot the average error """
	text = font.render(input, 1, (100, 100, 100))
	background.blit(text, (10, 10))

def plotLowerBar():
	""" Plot the toolbar """
	lbar= pygame.Surface((800, 50))
	lbar.fill((0, 0, 0))
	background.blit(lbar, (0, 600-50))
	
	rowNbr = 560
	# Save
	text = font.render("Save", 1, (200, 200, 200))
	background.blit(text, (10, rowNbr))
	
	# Paint
	if inputDict[PAINT]:
		text = font.render("Paint", 1, GREEN)
	else:
		text = font.render("Paint", 1, (200, 200, 200))
	background.blit(text, (100, rowNbr))
	
	# Get positions
	if inputDict[TIMES]:
		text = font.render("Times", 1, GREEN)
	else:
		text = font.render("Times", 1, (200, 200, 200))
	background.blit(text, (200, rowNbr))
	
	# Show true path
	if inputDict[INPUT]:
		text = font.render("Input", 1, GREEN)
	else:
		text = font.render("Input", 1, (200, 200, 200))
	background.blit(text, (300, rowNbr))
	
	# Show predicted
	if inputDict[PRED]:
		text = font.render("Pred", 1, GREEN)
	else:
		text = font.render("Pred", 1, (200, 200, 200))
	background.blit(text, (400, rowNbr))
	
	# Erase
	text = font.render("Erase", 1, (200, 200, 200))
	background.blit(text, (500, rowNbr))
	
	
	
def checkCommand(x):
	""" Check which commnad to execute """
	if x <= 70:
		return SAVE
	elif x > 70 and x <= 170:
		return PAINT
	elif x > 170 and x <= 270:
		return TIMES
	elif x > 270 and x <= 370:
		return INPUT
	elif x > 370 and x <= 470:
		return PRED
	elif x > 470 and x <= 570:
		return ERASE
	else:
		return UNKNOWN



# // ############################# GameObject   ###############################
class GameObject(pygame.sprite.Sprite):

	def __init__(self):
		""" Game object (not used) """
		#call Sprite initializer
		pygame.sprite.Sprite.__init__(self) 

# // ############################# PrintArray   ###############################
class PrintArray:
	def __init__(self, a, text = ""):
		""" Debug method """
		print "<-- Array: " + text + " -->"
		for x in range(0, len(a)):
			print a[x]

# // ############################# Path         ###############################
class Path:
	""" Defines a path for a virutal object """

	def __init__(self):
		""" Init the object """
		self.pairList = []

	def getCoordinates(self, vtime):
		""" Get the coordinates based on the actual time """
		i = 0 # Index counter for the pairList
		while i < len(self.pairList) and self.pairList[i][FIRSTELEMENT] <= vtime:
			i += 1
		if i == 0:
			return self.pairList[0][SECONDELEMENT]
		else:
			return self.pairList[i-1][SECONDELEMENT]

	def oldgetTime(self, vtime):
		""" Get the coordinates based on the actual time """
		i = 0 # Index counter for the pairList
		while i < len(self.pairList) and self.pairList[i][FIRSTELEMENT] <= vtime:
			i += 1
		return self.pairList[i-1][FIRSTELEMENT]

	def getPathObject(self, vtime):
		""" Get the pathObject based on the actual time """
		i = 0 # Index counter for the pairList
		while i < len(self.pairList) and self.pairList[i][FIRSTELEMENT] <= vtime:
			i += 1
		return self.pairList[i-1]

	def getTimeByCoordinates(self, pos):
		""" Get the time based on the actual coordinates """
		for x in self.pairList:
			if (pos[0] - 5) < int(x[1][0]) and (pos[0] + 5) > int(x[1][0]) and \
				(pos[1] - 5) < int(x[1][1]) and (pos[1] + 5) > int(x[1][1]):
				return x[0]

	def getMaxTime(self):
		""" Get the latest occurance in the path """
		lastTime = -1
		if  len(self.pairList) > 0:
			lastTime = self.pairList[-1]
		return lastTime[FIRSTELEMENT]

	def getLastObject(self):
		""" Get the lastest object in the path """
		lastTime = -1
		if  len(self.pairList) > 0:
			lastTime = self.pairList[-1]
		return lastTime

	def getFirstObject(self):
		""" Get the first object in the path """
		try:
			return self.pairList[0]
		except:
			print "FIX ...."
			sys.exit()


	def getAllObjects(self):
		""" Get all objects in the path """
		return self.pairList

	def getMeanDeltaT(self):
		""" Calculate the mean DT from all objects in the path """
		result = []
		a = self.pairList[:-1]
		b = self.pairList[1:]
		
		for t in range(len(a)):
			result.append(abs(b[t][TIME] - a[t][TIME]))
		return asarray(result).mean()

	def setPointPath(self, position):
		""" Add a point to the path """
		self.pairList.append((time.time(), array(position)))

	def clearPath(self):
		""" Clear the path """
		self.pairList = []

	def getSize(self):
		""" Return the size of the path """
		return len(self.pairList)

	def append(self, array):
		""" Append an path object to the path """
		for po in array:
			self.pairList.append(po)

	def getTime(self, vtime):
		""" Get the time that is closest to vtime """
		# Special written binary search for Path objects.
		try:
			# Get the length of the list and map to integers
			iFirst = 0
			iLast = len(self.pairList) - 1
			
			if vtime < self.pairList[iFirst][TIME]:
				return self.pairList[iFirst][TIME]
			
			if vtime > self.pairList[iLast][TIME]:
				return self.pairList[iLast][TIME]
			
			middle = -1
			while self.pairList[iFirst][TIME] <= self.pairList[iLast][TIME]:
				middle = (iFirst + iLast) / 2
				if vtime < self.pairList[middle][TIME]:
					iLast = middle - 1
				elif vtime > self.pairList[middle][TIME]:
					iFirst = middle + 1
				else:
					# If we found our element, then check which is best of this element and the last element
					if abs(self.pairList[middle][TIME] - vtime) < \
						abs(self.pairList[middle-1][TIME] - vtime):
						return self.pairList[middle][TIME]
					else:
						return self.pairList[middle-1][TIME]
			
			if abs(self.pairList[middle][TIME] - vtime) < \
				abs(self.pairList[middle-1][TIME] - vtime):
				return self.pairList[middle][TIME]
			else:
				return self.pairList[middle-1][TIME]
		except IndexError:
			if DEBUG:
				print "Indvalid index (out of index)"
	
	def setDefaultPath(self):
		""" Setting a straight line from x = (100, 540) and y = 240 withing 10 seconds """
		for i in range(0, 1000):
			t = 0.01 * i
			x = ((440.0 / 1000.0) * i) + 100
			y = MIDDLE
			self.pairList.append((t, array([x,y])))

	def printPath(self):	
		""" Prints a path """
		try:
			for k in range(len(self.pairList)):
					print "%6.3f , %s" % (self.pairList[k][FIRSTELEMENT], \
							self.pairList[k][SECONDELEMENT])
		except TypeError:
			if DEBUG:
				print "We got a type error in: %s" % self.pairList[k]

	def dumpPathToFile(self, filename):
		""" Write the path to a file """
		print "\nDumping paths to", filename
		if len(self.pairList) > 0:
			file = open(filename, 'w')
			firstTime = self.pairList[0][FIRSTELEMENT]
			for k in range(len(self.pairList)):
				line = str(self.pairList[k][FIRSTELEMENT] - firstTime) + \
						';' + \
						str(self.pairList[k][SECONDELEMENT][FIRSTELEMENT]) + \
						';' + \
						str(self.pairList[k][SECONDELEMENT][SECONDELEMENT]) + \
						'\n'
				file.write(line)
			file.close()
		print "\nDump succeded"
	
	def loadPathFromFile(self, filename):
		""" Load a path from a file """
		print "\nTry to load paths from: %s" % filename
		file = open(filename, 'r')
		for line in file.readlines():
			sline = line.split(";")
			vtime = float(sline[0])
			vx = float(sline[1])
			vy = float(sline[2])
			self.pairList.append((vtime, array([vx, vy])))
		file.close()
		print "\nPath successfully loaded from", filename

# // ############################# PixelPlotter ###############################
class PixelPlotter:
	""" Class that plot paths """
	def __init__(self):
		pass
	
	def plotLinedPath(self, path, color):
		""" Plots a lined path """
		try:
			if path.getSize() > 0:
				lastPoint = path.pairList[0][POINT]
				for k in range(len(path.pairList)):
					x = int(path.pairList[k][POINT][X])
					y = int(path.pairList[k][POINT][Y])
					pygame.draw.line(background, color, lastPoint, (x, y), 1)
					lastPoint = (x, y)
		except TypeError:
			if DEBUG:
				print "TypeError when trying to plot"

	def plotPath(self, path, color):
		""" Plot a path, only pixels """
		if path.getSize() > 0:
			for k in range(len(path.pairList)):
				x = int(round(path.pairList[k][POINT][X]))
				y = int(round(path.pairList[k][POINT][Y]))
				try:
					background.set_at((x, y), color)
				except OverflowError:
					if DEBUG:
						"Predicted very big values"
						print x, y

	def plotNetworkPoints(self, path, color, size=5):
		""" Plot points as circles """
		if path.getSize() > 0:
			for k in range(len(path.pairList)):
				x = int(round(path.pairList[k][POINT][X]))
				y = int(round(path.pairList[k][POINT][Y]))
				pygame.draw.circle(background, color, (x, y), size,  1)

# // ############################# PathManager  ###############################
class PathManager:
	""" Class that handles a set of paths """
	def __init__(self):
		""" Initialize the list """
		self.pathObjects = []

	def registerPath(self, path):
		""" Register a path """
		self.pathObjects.append(path)

	def eraseAllPaths(self):
		""" Clears the list """
		for p in self.pathObjects:
			p.clearPath()
			if DEBUG:
				print "\nAll paths has been cleared"

# // ############################# PathRandomizer    ##########################
class PathRandomizer:
	""" Generate a random path based on a true path """
	def __init__(self):
		pass

	def generatePathObjects(self, ipath, idelay, ijitter):
		""" Generate a random path based on a true path """
		gp = []
		maxTime = ipath.getMaxTime()
		totalNumberOfSteps = int(math.ceil(maxTime / idelay))
		for po in range(0, totalNumberOfSteps+1):
			timeForNextPathObject = (po * idelay) + (random.random() * ijitter)
			if timeForNextPathObject >  maxTime:
				gp.append(ipath.getLastObject())
				break
			else:
				gp.append(ipath.getPathObject(timeForNextPathObject))
		return gp

# // ############################# PathComparer      ##########################
class PathComparer:
	""" Class that compares two paths """
	
	def __init__(self):
		pass

	def compare(self, truePath, predictedPath):
		""" Do the comparision between two paths """
		distance = []
		firstOccuranceInPredictedPath = predictedPath.getFirstObject()[TIME]
		for po in truePath.getAllObjects()[5:]:
			ttime = po[TIME]
			if ttime > firstOccuranceInPredictedPath:
				tpos = po[POINT]
				ptime = predictedPath.getTime(ttime)
				ppos = predictedPath.getCoordinates(ttime)
				distance.append(getDistance(tpos, ppos))
		try:
			return asarray(distance).mean()
		except ZeroDivisionError:
			print "Zero Division error in PathComparer"
			sys.exit()

# // ############################# TimeDistributor   ############################
class TimeDistributor:
	""" Class that make use of old time values to predict new values """
	def __init__(self, workingValue=2):
		""" Initialize the object """
		self.timeList = []
		self.maxTimeUnit = -1
		self.workingValue = workingValue

	def addTime(self, t):
		""" Add a time """
		self.timeList.append(t)
		self.maxTimeUnit += 1


	def predictNextTimeValue(self):
		""" Compute the next predicted time occurance """
		resultList = []
		result = 0
		try:
			nbrOfTimeUnitsToUse = self.workingValue
			if nbrOfTimeUnitsToUse > self.maxTimeUnit:
				nbrOfTimeUnitsToUse = self.maxTimeUnit
				
			lastTime = self.timeList[-1]
			lastNTimes = self.timeList[self.maxTimeUnit - self.workingValue:]
			toTimes = lastNTimes[1:]
			fromTimes = lastNTimes[:-1]
			for i in range(len(toTimes)):
				resultList.append(toTimes[i] - fromTimes[i])
			
			if self.maxTimeUnit < self.workingValue:
				result = self.timeList[self.maxTimeUnit]
			else:
				result = asarray(resultList).mean() + lastTime
			return result
		except IndexError:
			print "\nNo time values to use"
			return -1

# // ############################# FIRExtrapolator ###########################
class FIRExtrapolator:
	""" Class that FIR-extrapolate """
	def __init__(self):
		pass

	def  extrapolate(self, vPositions, vCoefficients):
		""" Make a matrix multiplication, i.e. FIR extrapolate """
		try:
			return matrixmultiply(vPositions, vCoefficients)
		except ValueError:
			print "Dimensions isn't equal: a = %d  b = %d" % \
				(len(vPositions), len(vCoefficients))

# // ############################# Quadratic Prediction ##########################
class QuadraticExtrapolator:
	""" Extrapolator that use quadratic prediction """
	
	def __init__(self):
		""" Initialize the object """
		self.aPath = []
		self.td = TimeDistributor()
		self.name = "Quadratic (FIR)"

	def setTimeHistoryUsage(self, n):
		""" Set how many points of time history it should use """
		self.td.workingValue = n

	def getTimeHistoryUsage(self):
		""" Get how many points of time history it use """
		return self.td.workingValue

	def extrapolate(self, ipath):
		""" Do the extrapolation, step 1 """
		if ipath.getSize() > 3:
			# Make a copy of list, shifted one step
			pathObjects = ipath.getAllObjects()
			pO = pathObjects[:-2]
			qO = pathObjects[1:-1]
			rO = pathObjects[2:]
				
			self.td.addTime(pO[0][TIME])
			self.td.addTime(qO[0][TIME])
			
			for i in range(len(pO)):
				# 1. Add time of the last point all the time
				self.td.addTime(rO[i][TIME])
				# Get some kind of FIR extrapolation
				self.aPath.append(self.pointExtrapolate(pO[i], qO[i], rO[i]))
			return self.aPath

	def pointExtrapolate(self, p, q, r):
		""" Predict by using points, step 2 """
		# Predict x- and y-coordinate
		predictedTime = self.td.predictNextTimeValue()
		x = self.vectorExtrapolate(p[POINT][X], q[POINT][X], r[POINT][X], predictedTime, r[TIME])
		y = self.vectorExtrapolate(p[POINT][Y], q[POINT][Y], r[POINT][Y], predictedTime, r[TIME])
		result = predictedTime, array([x, y])
		return result

	def vectorExtrapolate(self, p, q, r, t, tr):
		""" Extrapolate individual vectors, step 3 """
		try:
			dt = t - tr
			c0 = 1 + dt + 0.5*(dt**2)
			c1 = -dt - (dt**2)
			c2 = 0.5 * (dt**2)
			return r + (matrixmultiply([c0, c1, c2], [r, q, p]) - r) / dt
		except ValueError:
			print "Value error in Quadratic"
			sys.exit()
		except ZeroDivisionError:
			if DEBUG:
				print "Dividing by zero: tp = tq = %.3f" % (t-tr)
			return r

	def clear(self):
		""" Clear the predictor """
		self.aPath = []
		self.td = TimeDistributor(self.getTimeHistoryUsage())

# // ############################# Catmull-Rom Prediction ##########################
class CatmullRomExtrapolator:
	""" Extrapolator that use Catmull-Rom prediction """
	
	def __init__(self):
		""" Initialize the object """
		self.aPath = []
		self.td = TimeDistributor()
		self.name = "Catmull-Rom(FIR)"

	def setTimeHistoryUsage(self, n):
		""" Set how many points of time history it should use """
		self.td.workingValue = n

	def getTimeHistoryUsage(self):
		""" Get how many points of time history it use """
		return self.td.workingValue

	def extrapolate(self, ipath):
		""" Do the extrapolation, step 1 """
		if ipath.getSize() > 3:
			# Make a copy of list, shifted one step
			pathObjects = ipath.getAllObjects()
			pO = pathObjects[:-2]
			qO = pathObjects[1:-1]
			rO = pathObjects[2:]
				
			self.td.addTime(pO[0][TIME])
			self.td.addTime(qO[0][TIME])
			
			for i in range(len(pO)):
				# 1. Add time of the last point all the time
				self.td.addTime(rO[i][TIME])
				# Get some kind of FIR extrapolation
				self.aPath.append(self.pointExtrapolate(pO[i], qO[i], rO[i]))
			return self.aPath

	def pointExtrapolate(self, p, q, r):
		""" Predict by using points, step 2 """
		# Predict x- and y-coordinate
		predictedTime = self.td.predictNextTimeValue()
		x = self.vectorExtrapolate(p[POINT][X], q[POINT][X], r[POINT][X])
		y = self.vectorExtrapolate(p[POINT][Y], q[POINT][Y], r[POINT][Y])
		result = predictedTime, array([x, y])
		return result

	def vectorExtrapolate(self, p, q, r):
		""" Extrapolate individual vectors, step 3 """
		try:
			# FIR coefficients
			c0 = 2.5
			c1 = -2
			c2 = 0.5
			return matrixmultiply([c0, c1, c2], [r, q, p])
		except ValueError:
			print "Value error in Catmull-Rom extrapolator"
			sys.exit()
		except ZeroDivisionError:
			if DEBUG:
				pass
				# This is impossible
				print "Dividing by zero (Catmull): dt: %.3f" % dt
			return r

	def clear(self):
		""" Clear the predictor """
		self.aPath = []
		self.td = TimeDistributor(self.getTimeHistoryUsage())

# // ############################# Richardson Prediction ##########################
class RichardsonExtrapolator:
	""" Extrapolator that use Richardson Extrapolator prediction """
	
	def __init__(self):
		""" Initialize the object """
		self.aPath = []
		self.td = TimeDistributor()
		self.name = "Richardson (FIR)"

	def setTimeHistoryUsage(self, n):
		""" Set how many points of time history it should use """
		self.td.workingValue = n

	def getTimeHistoryUsage(self):
		""" Get how many points of time history it use """
		return self.td.workingValue

	def extrapolate(self, ipath):
		""" Do the extrapolation, step 1 """
		if ipath.getSize() > 3:
			# Make a copy of list, shifted one step
			pathObjects = ipath.getAllObjects()
			pO = pathObjects[:-2]
			qO = pathObjects[1:-1]
			rO = pathObjects[2:]
				
			self.td.addTime(pO[0][TIME])
			self.td.addTime(qO[0][TIME])
			
			for i in range(len(pO)):
				# 1. Add time of the last point all the time
				self.td.addTime(rO[i][TIME])
				# Get some kind of FIR extrapolation
				self.aPath.append(self.pointExtrapolate(pO[i], qO[i], rO[i]))
			return self.aPath

	def pointExtrapolate(self, p, q, r):
		""" Predict by using points, step 2 """
		# Predict x- and y-coordinate
		predictedTime = self.td.predictNextTimeValue()
		x = self.vectorExtrapolate(p[POINT][X], q[POINT][X], r[POINT][X], predictedTime, p[TIME], q[TIME], r[TIME])
		y = self.vectorExtrapolate(p[POINT][Y], q[POINT][Y], r[POINT][Y], predictedTime, p[TIME], q[TIME], r[TIME])
		result = predictedTime, array([x, y])
		return result

	def vectorExtrapolate(self, p, q, r, t, tp, tq, tr):
		""" Extrapolate individual vectors, step 3 """
		try:
			t0t1 = tr - tq
			t0t2 = tr - tp
			dt1 = t - tq
			
			c0 = (4.0 / (3*t0t1) - (1.0 / (3*t0t2))) * dt1
			c1 = -((4.0*dt1) / (3.0*t0t1)) + 1
			c2 = dt1 / (3.0 * t0t2)
			return matrixmultiply([c0, c1, c2], [r, q, p])
		except ValueError:
			print "Value error in Richardsonextrapolator"
			sys.exit()
		except ZeroDivisionError:
			if DEBUG:
				pass
				print "Dividing by zero: t0t1: %.3f, t0t2: %.3f, dt1: %.3f" % (t0t1, t0t2, dt1)
			return r

	def clear(self):
		""" Clear the predictor """
		self.aPath = []
		self.td = TimeDistributor(self.getTimeHistoryUsage())

# // ############################# Linear Prediction ##########################
class LinearExtrapolator:
	""" Extrapolator that use Linear prediction """
	
	def __init__(self):
		""" Initialize the object """
		self.linearPath = []
		self.td = TimeDistributor()
		self.name = "Linear (FIR)"

	def setTimeHistoryUsage(self, n):
		""" Set how many points of time history it should use """
		self.td.workingValue = n

	def getTimeHistoryUsage(self):
		""" Get how many points of time history it use """
		return self.td.workingValue

	def extrapolate(self, ipath):
		""" Do the extrapolation, step 1 """
		if ipath.getSize() > 2:
			# Make a copy of list, shifted one step
			pathObjects = ipath.getAllObjects()
			toPath = pathObjects[1:]
			fromPath = pathObjects[:-1]
			
			# Doesn't matter if we use length of toPath or fromPath, they should be equal
			for i in range(len(toPath)):
				self.linearPath.append(self.pointExtrapolate(toPath[i], fromPath[i]))
			return self.linearPath

	def pointExtrapolate(self, p, q):
		""" Predict by using points, step 2 """
		predictedTime = -1
		x = -1
		y = -1
		# 1. Predict the time when next event occur
		self.td.addTime(p[TIME])
		predictedTime = self.td.predictNextTimeValue()
		# 2. Predict x- and y-coordinate
		x = self.vectorExtrapolate(p[POINT][X], q[POINT][X], predictedTime, \
			p[TIME], q[TIME])
		y = self.vectorExtrapolate(p[POINT][Y], q[POINT][Y], predictedTime, \
			p[TIME], q[TIME])
		result = predictedTime, array([x, y])
		return result

	def vectorExtrapolate(self, p, q, t, tp, tq):
		""" Extrapolate individual vectors, step 3 """
		try:
			return (p + ((t-tp)/(tp-tq))*(p-q))
		except ZeroDivisionError:
			if DEBUG:
				print "Dividing by zero: tp = tq = %.3f" % tp
			return p

	def clear(self):
		""" Clear the predictor """
		self.linearPath = []
		self.td = TimeDistributor(self.getTimeHistoryUsage())

# // ############################# Linear Prediction ##########################
class LinearInterpolator:
	""" Class that interpolates a path by using linear interpolation """
	
	def __init__(self):
		""" Initialize the object """
		self.ip = []
		self.name = "linear"

	def interpolate(self, ipath):
		""" Do the interpolation, step 1 """
		# Make a copy of list, shifted one step
		pathObjects = ipath.getAllObjects()
		toPath = pathObjects[1:]
		fromPath = pathObjects[:-1]
		
		for i in range(len(toPath)):
			self.pointInterpolate(fromPath[i], toPath[i])
		
		return self.ip

	def pointInterpolate(self, q, p):
		""" Interpolate pointwise, step 2 """
		q_time = q[TIME]
		p_time = p[TIME]
		
		qx = q[POINT][X]
		px = p[POINT][X]
		
		qy = q[POINT][Y]
		py = p[POINT][Y]
		
		nbrOfSteps = math.ceil(max(abs(py-qy), abs(px-qx), 1))
		
		interPolatedTimes = self.vectorInterpolate(q_time, p_time, nbrOfSteps)
		interPolatedQ = self.vectorInterpolate(qx, px, nbrOfSteps)
		interPolatedP = self.vectorInterpolate(qy, py, nbrOfSteps)
		for po in zip(interPolatedTimes, array(zip(interPolatedQ, interPolatedP))):
			self.ip.append(po)

	def vectorInterpolate(self, a, b, nbrOfSteps):
		""" Interpolate vectorwise, step 3 """
		result = []
		for i in range(int(nbrOfSteps) + 1):
			t = float(i / nbrOfSteps)
			result.append(a + (b-a)*t)
		return result

	def clear(self):
		""" Clear the interpolator """
		self.ip = []

# // ############################# Catmull-Rom Interpolation ##########################
class CatmullRomInterpolator:
	""" Class that interpolates a path by using Catmull-Rom interpolation """
	
	def __init__(self):
		""" Initialize the object """
		self.ip = []
		self.name = "Catmull-Rom"
		self.tau = 0.7
		self.M = array([\
			[0, 1, 0, 0],\
			[-self.tau, 0,self.tau, 0],\
			[2*self.tau, self.tau-3, 3-2*self.tau, -self.tau],\
			[-self.tau, 2-self.tau, self.tau-2, self.tau]])

	def interpolate(self, ipath):
		""" Do the interpolation, step 1 """
		# Make a copy of list, shifted one step
		pathObjects = ipath.getAllObjects()
		pO = pathObjects[:-2]
		qO = pathObjects[1:-1]
		rO = pathObjects[2:]
		
		for i in range(len(pO)):
			try:
				self.pointInterpolate(pO[i], qO[i], rO[i], rO[i+1])
			except IndexError:
				if DEBUG:
					print "\nIndex error, CMRI, i =", i
		
		return self.ip

	def pointInterpolate(self, p, q, r, pred):
		""" Interpolate pointwise, step 2 """
		p_time = p[TIME]
		q_time = q[TIME]
		r_time = r[TIME]
		pred_time = pred[TIME]
		
		qx = q[POINT][X]
		px = p[POINT][X]
		rx = r[POINT][X]
		predx = pred[POINT][X]
		
		qy = q[POINT][Y]
		py = p[POINT][Y]
		ry = r[POINT][Y]
		predy = pred[POINT][Y]
		
		x_vect = array([px, qx, rx, predx])
		y_vect = array([py, qy, ry, predy])
		t_vect = array([p_time, q_time, r_time, pred_time])
		
		step = max(abs(rx-qx), abs(ry-qy))
		
		try:
			if step < 1:
				step = 1
		except ZeroDivisionError:
			step = 1
			if DEBUG:
				print "cubicInterpolation, stepsize =", step
		
		self.cubicInterpolation(x_vect, y_vect, t_vect, step)

	def cubicInterpolation(self, vX, vY, vT, stepsize=100):
		""" Interpolate vectorwise, step 3 """
		for i in arange(0.0, 1.0, 1.0/stepsize):
			M = array([\
				[0, 1, 0, 0],\
				[-self.tau, 0,self.tau, 0],\
				[2*self.tau, self.tau-3, 3-2*self.tau, -self.tau],\
				[-self.tau, 2-self.tau, self.tau-2, self.tau]])
			u = array([1, i, i**2, (i**2)*i])
			uM = matrixmultiply(u, M)
			x = matrixmultiply(uM,  transpose(vX))
			y = matrixmultiply(uM,  transpose(vY))
			t = matrixmultiply(uM,  transpose(vT))
			self.ip.append((t, array([x, y])))

	def clear(self):
		""" Clear the interpolator """
		self.ip = []

# // ############################# Simulator         ##########################
class Simulator:
	""" Class that is responsible for doing the simulation """
	
	def __init__(self, truePath, extraPolator, interPolator, nbrOfSimulations=100, mode=0):
		""" Initilize objects, paths, extrapolator and interpolator """
		self.tp = truePath
		self.ep = extraPolator
		self.ip = interPolator
		self.nbrSim = nbrOfSimulations
		self.endResult = []
		self.dt = 0.5
		self.jitter = 0.1
		self.lastPath = None
		self.lastSimulated = None
		self.lastExtrapolated = None
		self.fileName = "paintPath.txt"
		self.mode = mode

	def doSimulation(self):
		""" Do the actual simulation """
		lastTime = time.time()
		for x in range(0, self.nbrSim):
			pRandomizer = PathRandomizer()
			extraPolatedPath = Path()
			interPolatedPath = Path()
			simulatedNetworkUpdates = Path()
			pc = PathComparer()

			# We put these newly generated update states into a Path Object.
			# We get a random array of path objects, i.e. we simulate updates from a network.
			simulatedNetworkUpdates.append(pRandomizer.generatePathObjects(self.tp, self.dt, self.jitter))
			
			# And put the path into an extrapolator.
			# Append the extrapolated posistions to a Path object.
			extraPolatedPath.append(self.ep.extrapolate(simulatedNetworkUpdates))
			
			if self.mode == 0:
				# We interpolate between the extrapolated positions.
				# and put the points into a Path object.
				interPolatedPath.append(self.ip.interpolate(extraPolatedPath))
				
				# And compare the result of the true path and the interpolated path.
				result = pc.compare(self.tp, interPolatedPath)
				self.endResult.append(result)
				self.lastPath = interPolatedPath
			else:
				result = pc.compare(extraPolatedPath, self.tp)
				self.endResult.append(result)
				self.lastPath = self.tp
			
			print "<-- Average error: %.3f -->" % result
			
			if x % 10 == 0:
				newTime = time.time()
				print "Simulating ... %6.3f seconds" % (newTime - lastTime)
				lastTime = newTime
				
			self.lastPath = interPolatedPath
			self.lastSimulated = simulatedNetworkUpdates
			self.lastExtrapolated = extraPolatedPath
			self.reset()

		meanResult = array(self.endResult).mean()
		# Save the result to a file 
		file = open("results.txt", 'a')
		line = str("%s" % self.fileName) + \
					';' + \
					str(self.ep.name) + \
					';' + \
					str(self.ip.name) + \
					';' + \
					str("%.4s;%.4s;%.4s;%d" % (self.nbrSim, self.dt, self.jitter, self.ep.getTimeHistoryUsage())) + \
					';' + \
					str("%6.3f" % meanResult) + \
					'\n'
		file.write(line)
		file.close()
		return meanResult

	def reset(self):
		""" Reset the interpolator and extrapolator """
		self.ip.clear()
		self.ep.clear()

	def setConstants(self, dt, jitter):
		""" Set DT and jitter """
		self.dt = dt
		self.jitter = jitter

	def setFileName(self, fn):
		""" Set the filename """
		self.fileName = fn

	def getLastPath(self):
		""" Get the last path that was simulated """
		return self.lastPath

	def getLastSimulated(self):
		""" Get the last path that was simulated """
		return self.lastSimulated

	def getLastExtrapolatedPath(self):
		""" Get the last extrapolated path that was simulated """
		return self.lastExtrapolated


# // ############################# Main program ###############################
def main():
	global paintMode
	global timeMode
	global showFilePlot
	global showFilePaintPathPoints
	global showPredictedPlot
	global showExtrapolatedPathPoints
	global showNetworkPoints
	global clearScreen
	global getValue
	
	# Name of the filename that has the path that we would simulate
	fileName = "paintPath.txt"
	if len(sys.argv) == 2:
		fileName = sys.argv[1]

	print "\nSetting up simulation"
	
	# Misc inits
	running = TRUE
	t0 = time.time()
	pp = PixelPlotter()
	lastMousePosition = (0, 0)
	
	# Paths
	paintPath = Path()
	filePaintPath = Path()
	trueInterpolatedInputPath = Path()
	
	# Pathmanager
	pm = PathManager()
	
	nbrOfSimulations = 1
	meanStepFactor = 1
	useJitter = 0
	timeHistory = 1
	
	# Extrapolators
	extrapolator = LinearExtrapolator()
	#extrapolator = QuadraticExtrapolator()
	extrapolator = RichardsonExtrapolator()
	#extrapolator = CatmullRomExtrapolator()
	extrapolator.setTimeHistoryUsage(1)
	
	# Interpolators
	interpolator = LinearInterpolator()
	#interpolator = CatmullRomInterpolator()

	
	# Register Paths
	pm.registerPath(paintPath)
	pm.registerPath(filePaintPath)
	pm.registerPath(trueInterpolatedInputPath)
	
	# Load the file
	try:
		filePaintPath.loadPathFromFile(fileName)
	except IOError:
		print "The file does not exist: %s" % fileName
		sys.exit(1)
			
	# Intepolate between the points in the true path, so we get a continous path on the screen.
	trueInterpolatedInputPath.append(LinearInterpolator().interpolate(filePaintPath))
	interpolator.clear()
			
	meanDT = filePaintPath.getMeanDeltaT()
	print "Mean DT is: %.3f" % meanDT
	
	# Clamp step so we have some steps
	if filePaintPath.getSize() < (meanDT * meanStepFactor):
		meanStepFactor = 1
	latency = meanStepFactor * meanDT
	jitter = meanStepFactor * meanDT*useJitter
	if jitter == 0:
		# No need to do more simulations if we don't have any jitter
		nbrOfSimulations = 1
	print "Using latency: %6.3f and jitter: %6.3f" % (latency, jitter)
	
	# Simulates between interpolated true path and extrapolated path.
	simulator = Simulator(trueInterpolatedInputPath, extrapolator, interpolator, nbrOfSimulations)
	simulator.setConstants(latency, jitter)
	simulator.setFileName(str(fileName))
	
	plotText(str("Average error is: %6.3f" % (simulator.doSimulation())))
	plotText(str("Extrapolator: %s     Interpolator: %s" % \
		(extrapolator.name, interpolator.name)), 20)
	while running:
		for event in pygame.event.get():
			if event.type == MOUSEMOTION:
				if paintMode:
					pos = pygame.mouse.get_pos()
					pygame.draw.line(background, BLUE, lastMousePosition, pos, 1)
					lastMousePosition = pos
					paintPath.setPointPath(pos)
				if timeMode:
					pos = pygame.mouse.get_pos()
					result = simulator.getLastPath().getTimeByCoordinates(pos)
					truePos = trueInterpolatedInputPath.getCoordinates(result)
					dist = getDistance(pos, truePos)
					if result:
						background.fill((250, 250, 250))
						redPos = str("(%s, %s)" % (round(pos[0]), round(pos[1])))
						bluePos = str("(%s, %s)" % (round(truePos[0]), round(truePos[1])))
						plotText(str("Red dot: %s" % redPos), 30)
						plotText(str("Blue dot: %s" % bluePos), 60)
						plotText(str("Time: %.3fs    Distance: %.3f" % (result, dist)))
						pygame.draw.circle(background, RED, (int(round(pos[0])), int(round(pos[1]))), 5,  1)
						pygame.draw.circle(background, BLUE, (int(round(truePos[0])), int(round(truePos[1]))), 5,  1)
					
			elif event.type == MOUSEBUTTONDOWN:
				pos = pygame.mouse.get_pos()
				# Check if it's a click in command toolbar
				if pos[Y] > 550:
					cmd = checkCommand(pos[X])
					inputDict[cmd] = not inputDict[cmd]
					disableExcept(cmd)
					
					if cmd == ERASE:
						clearScreen = TRUE
						paintMode = FALSE
						showFilePlot = FALSE
						showPredictedPlot = FALSE
						showNetworkPoints = FALSE
					elif cmd == SAVE:
						paintPath.dumpPathToFile("paintPath.txt")
						plotText("Dumped painted path to file: paintPath.txt")
						
				else:
					if inputDict[PAINT]:
						clearScreen = paintMode = not paintMode
						print "PaintMode: " + ONOFFTEXT[paintMode]
						lastMousePosition = pygame.mouse.get_pos()
					elif inputDict[TIMES]:
						timeMode = not timeMode
						print "Time mode: " + ONOFFTEXT[timeMode]


			# Keyboard handling
			elif event.type in (QUIT, KEYDOWN):
				if event.key in (K_ESCAPE, K_q):
					sys.exit()
				elif event.key == K_1:
					showFilePlot = not showFilePlot
					clearScreen = not showFilePlot
					print "Show true path: " + ONOFFTEXT[showFilePlot]
				elif event.key == K_2:
					showPredictedPlot = not showPredictedPlot
					clearScreen = not showPredictedPlot
					print "Show predicted path: " + ONOFFTEXT[showPredictedPlot] 
				elif event.key == K_3:
					showNetworkPoints = not showNetworkPoints
					clearScreen = not showNetworkPoints
					print "Show network points: " + ONOFFTEXT[showNetworkPoints]
				elif event.key == K_4:
					showFilePaintPathPoints = not showFilePaintPathPoints
					clearScreen = not showFilePaintPathPoints
					print "Show true points: " + ONOFFTEXT[showFilePaintPathPoints]
				elif event.key == K_5:
					showExtrapolatedPathPoints = not showExtrapolatedPathPoints
					clearScreen = not showExtrapolatedPathPoints
					print "Show extrapolated points: " + ONOFFTEXT[showExtrapolatedPathPoints]	
		
		# Clear the screen
		if clearScreen:
			if DEBUG:
				print "Clearing the screen"
			background.fill((250, 250, 250))
			screen.blit(background, (0, 0))
			clearScreen = FALSE

		# If we are in paint mode we won't allow anything else to be drawn to screen.
		if showFilePlot and not inputDict[PAINT]:
			pp.plotPath(trueInterpolatedInputPath, BLUE)
			
		if showFilePaintPathPoints and not inputDict[PAINT]:
			pp.plotNetworkPoints(filePaintPath, BLUE, 3)
		
		# If we are in paint mode we won't allow anything else to be drawn to screen.
		if showPredictedPlot and not inputDict[PAINT]:
			pp.plotPath(simulator.getLastPath(), RED)
		
		if showExtrapolatedPathPoints and not inputDict[PAINT]:
			pp.plotNetworkPoints(simulator.getLastExtrapolatedPath(), RED, 3)
		
		# If we are in paint mode we won't allow anything else to be drawn to screen.		
		if showNetworkPoints and not inputDict[PAINT]:
			pp.plotNetworkPoints(simulator.getLastSimulated(), BLACK)

		plotLowerBar()
		screen.blit(background, (0, 0))
		pygame.display.flip()
		pygame.event.pump() 


if __name__ == '__main__':
	main() 
